/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import invariant from 'evojs/lib/invariant';

import StoreManager from './services/StoreManager';
import NetworkStatus from './services/NetworkStatus';
import SecurityToken from './services/SecurityToken';
import RequestSender from './services/requestSender/RequestSender';
import EntityPersistor from './services/EntityPersistor';
import Faussaire from 'faussaire';

import adaptFaussaireToRequestSender from './services/requestSender/adapter/faussaireAdapter';
import adaptFetchToRequestSender from './services/requestSender/adapter/fetchAdapter';

export default function ShaaraReactNativeTask() {
  return {
    /**
     * Return default services
     * @param register
     * @param parameters
     * @param env
     */
    registerServices(register, parameters, env) {
      let faussaire = null;
      const requestSenderImplementation = (() => {
        if(env === "test"){
          faussaire = Faussaire.create();
          return adaptFaussaireToRequestSender(faussaire);
        }

        // By default we rely on global.fetch
        // If there's an user-defined fetch function, we can use it
        if(parameters.fetch)
          return adaptFetchToRequestSender(parameters.fetch);

        invariant(
          typeof fetch === "function",
          "fetch is not defined. Cannot initialize the request sender."
        );

        return adaptFetchToRequestSender(fetch);
      })();

      register("store", StoreManager.create());
      register("network.status", NetworkStatus.getInstance());
      register("request.sender", RequestSender.createFactory(requestSenderImplementation));
      register("security.token", SecurityToken.create());
      register("entity.persistor", EntityPersistor.createFromContainer);

      if(faussaire) {
        register("faussaire", faussaire);
      }
    },

    /**
     * Allows to configure the store and the securityToken
     * @param config
     * @param kernel
     * @returns {Promise.<void>}
     */
    async configureKernel(config, kernel) {
      if(typeof config.getStoreConfiguration === "function") {
        await kernel.get("store").init(config.getStoreConfiguration());
      }
      if(typeof config.createStore === "function") {
        await kernel.get("store").init(config.getStoreConfiguration(), config.createStore);
      }

      if(typeof config.getTokenImplementation === "function") {
        kernel.get("security.token").setImplementation(config.getTokenImplementation(kernel._services));
      }

      if(typeof config.getEntityPersistorHandlers === "function") {
        const handlers = config.getEntityPersistorHandlers();
        Object.keys(handlers).forEach(key => {
          kernel.get("entity.persistor").addHandler(key, handlers[key]);
        })
      }
    }
  }
}