// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

type Action = {[key: string] : Object };
type Handler = (data : Action, context? : any) => { type: string }
type HandlerListener = { name: string, handle: Handler };

/**
 * Manage persisting data into the store using the mechanism of handlers.
 * Data given to the EntityPersistor must be normalized, that is, an entity doesn't have nested entities
 * inside. An object must follow this pattern :
 * {
 *   users: {
 *     "1": {
 *       id: 1,
 *       username: "foo",
 *     }
 *  },
 *  posts: {
 *    "2": {
 *      id: 2,
 *      userId: 1,
 *      title: "foo"
 *    }
 * }
 *
 * The Persistor will read the name of every properties of the object and will try to match them with handlers.
 * If one handler correspond, it will call it.
 * A handler is responsible for transforming data into a Redux action.
 * Then, the store method will dispatch all the actions.
 *
 * @param store
 * @constructor
 */
export function EntityPersistor(store : Object) {
  const handlers : Array<HandlerListener> = [];

  /**
   * Adds a handler
   * @param name
   * @param handle
   */
  this.addHandler = function(name : string, handle : Handler) : void {
    handlers.push({
      name,
      handle,
    });
  };

  /**
   * Return the handler matching the name
   * @param name
   * @returns {HandlerListener}
   */
  this.getHandler = function(name : string) : ?HandlerListener {
    return handlers.find((value : HandlerListener) => value.name === name);
  };

  /**
   * Return an array of actions to perform
   * @param data
   * @param context
   * @returns {Array.<*>}
   */
  this.getActions = function(data : Object, context : any) : Array<Action> {
    return Object.keys(data).map(key => {
      const handler = this.getHandler(key);
      if(handler === undefined) {
        // $FlowFixMe
        return undefined;
      }

      return handler.handle(data[key], context);
    }).filter(action => action !== undefined);
  };

  /**
   * Dispatch actions to store the data
   * @param data
   * @param context
   */
  this.store = function(data : Object, context : any) : void {
    store.dispatch(this.getActions(data, context));
  }
}

export default {
  createFromContainer(container : Object) : EntityPersistor {
    return new EntityPersistor(container.get("store"));
  }
}