/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import invariant from 'evojs/lib/invariant';

export type SecurityTokenType<$TokenType : mixed> = {
  setToken: (token: $TokenType) => void,
  getToken: () => $TokenType,
  getAppUserId: () => string,
};

/**
 * Provides an interface around a SecurityToken system.
 * We leave the implementation to the user so he can decide from which source to fetch data. In production it will
 * usually be the store.
 *
 * @constructor
 */
export function SecurityToken(){
  this.implementation = null;
}

/**
 * Throws an error if the implementation is missing or incorrect
 */
SecurityToken.prototype.assertIsImplemented = function(){
  invariant(
    typeof this.implementation === "object"
    && this.implementation !== null
    && typeof this.implementation.setToken === "function"
    && typeof this.implementation.getToken === "function",
    "You must provide a valid implementation to SecurityToken."
  )
};

/**
 * Update the implementation
 * @param impl
 */
SecurityToken.prototype.setImplementation = function(impl){
  this.implementation = impl;
};

/**
 * Set the token
 * @param token
 */
SecurityToken.prototype.setToken = function(token){
  this.assertIsImplemented();
  this.implementation.setToken(token);
};

/**
 * Return the token
 * @returns {$TokenType|*}
 */
SecurityToken.prototype.getToken = function(){
  this.assertIsImplemented();
  return this.implementation.getToken();
};

export default {
  /**
   * @returns {SecurityTokenType}
   */
  create() : SecurityTokenType {
    return new SecurityToken();
  }
}