// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import invariant from 'evojs/lib/invariant';
import { createStore, combineReducers, compose } from 'redux';
import reduxBatch from "../util/reduxBatch";

export type ReducerType = (state : ?mixed, action : { type: mixed }) => mixed;

/**
 * Manages the redux store
 * @constructor
 */
function StoreManager(){
  this._store = null;
  this._isConfigured = false;
}

export type StoreConfiguration = {
  reducers?: {[key: string]: ReducerType },
  debounce: number,
};

const defaultInitConfiguration : StoreConfiguration = {
  debounce: 500,
};

/**
 * Creates a reducer from nested objects
 * @param reducers
 * @returns {Reducer<any>}
 */
function assembleReducers(reducers : {[key: string] : Function | Object }) : {[key: string] : Function} {
  let reducerObj = {};
  for(let key in reducers) {
    if(reducers.hasOwnProperty(key) === false){
      continue;
    }

    if(typeof reducers[key] === "object"){
      reducerObj[key] = assembleReducers(reducers[key]);
    } else {
      reducerObj[key] = reducers[key];
    }
  }

  return combineReducers(reducerObj);
}

function defaultStoreFactory (config) {
  return createStore(config.reducers, compose(reduxBatch));
}
/**
 * Creates the store and rehydrates it
 *
 * @param config
 * @param storeFactory
 * @returns {Promise}
 */
StoreManager.prototype.init = async function(config : StoreConfiguration, storeFactory){
  config = Object.assign({}, defaultInitConfiguration, config);
  config.reducers = assembleReducers(config.reducers);

  invariant(
    config.reducers,
    "You must define a reducer",
  );

  storeFactory = typeof storeFactory === "function" ? storeFactory : defaultStoreFactory;
  this._store = await storeFactory(config);
  this._isConfigured = true;

  invariant(
    typeof this._store === "object" && this._store !== null &&
    typeof this._store.dispatch === "function" && typeof this._store.getState === "function",
    "the storeFactory function must return a valid store object"
  );

  return this._store;
};

/**
 * Throws an error if the store is not configured
 */
StoreManager.prototype.assertStoreIsConfigured = function(){
  invariant(
    this.isConfigured() === true,
    "You cannot access any property of the store as it is not configured yet. " +
    "Please call the init function to configure the store."
  )
};

/**
 * Return the state
 * @returns {S}
 */
StoreManager.prototype.getState = function(){
  this.assertStoreIsConfigured();
  return this._store.getState();
};

/**
 * Dispatches an action to the store
 * @returns {*<S>}
 */
StoreManager.prototype.dispatch = function(action){
  this.assertStoreIsConfigured();
  return this._store.dispatch(action);
};

/**
 * Returns true if the store is configured
 * @returns {boolean}
 */
StoreManager.prototype.isConfigured = function() : boolean {
  return this._isConfigured;
};

/**
 * Return the store
 * @returns {Store<S>|*|null}
 */
StoreManager.prototype.getStore = function(){
  return this._store;
};

export default {
  create(){
    return new StoreManager();
  }
}