// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import interpretResult from './interpretResult';
import Result from './Result';

import type {
  RequestConfiguration,
  ResultType,
  NetworkInfoType,
} from './definitionTypes';

export type FetchType = (request: RequestConfiguration) => Promise<ResultType>

/**
 * List of possible contentTypes
 * @type {{json: string, urlEncoded: string, formData: string}}
 */
export const ContentTypes = {
  json: "json",
  urlEncoded: "url-encoded",
  formData: "form-data",
};

/**
 * Noop returning true
 * @type {{isConnected: (function())}}
 */
const defaultNetworkInfo : NetworkInfoType = {
  isConnected(){
    return true;
  }
};

/**
 * Factory for the RequestSender
 * @constructor
 */
function RequestSenderFactory(){}
RequestSenderFactory.create = function(
  fetch: FetchType,
  networkInfo : NetworkInfoType = defaultNetworkInfo
) {
  return new RequestSender(fetch, networkInfo.isConnected);
};

/**
 * Creates a RequestSender factory based on a default implementation.
 * @param implementation
 */
RequestSenderFactory.createFactory = (implementation) => (services) => {
  return RequestSenderFactory.create(implementation, services.get("network.status"));
};

/**
 * @class RequestSender
 * @description
 * HttpRequestSenderFactory allow to send HTTP requests without caring about the implementation by providing
 * a simple interface for both requests and response.
 *
 * An HTTP implementation do the actual work. It must implement the HttpRequestImplement interface.
 */
export function RequestSender(fetch: FetchType, isConnected: () => boolean) {
  this.fetch = fetch;

  /**
   * Return true if the system is considered to be offline
   */
  this.isDisconnected = () => typeof isConnected === "function" && isConnected() === false;

  /**
   * List of middlewares
   * A middleware is called before actually calling the implementation. It allow further configuration
   * of the config object.
   *
   * @type {Array}
   */
  const middlewares = [];

  /**
   * Adds a middleware
   * @param middleware
   */
  this.addMiddleware = function(middleware : (config?: RequestConfiguration) => any) {
    middlewares.push(middleware);
  };

  /**
   * Call the middleware with the config
   * @param config
   */
  this.callMiddlewares = function(config : RequestConfiguration) {
    middlewares.forEach(middleware => middleware(config));
  };
}

/**
 * Create a valid configuration from the request configuration
 * @param config
 * @returns {RequestConfiguration}
 */
RequestSender.prototype.createConfiguration = function createConfiguration(config : RequestConfiguration) : RequestConfiguration {
  return {
    query: config.query || null,
    body: config.body || null,
    headers: config.headers || null,
    timeout: config.timeout || null,
    url: config.url,
    method: config.method,
    contentType: config.contentType || null,
    ...(config.payload ? { payload: config.payload } : {})
  };
};

/**
 * Perform the request
 * @param config
 * @returns {ResultType}
 */
RequestSender.prototype.send = function(config: RequestConfiguration): Promise<ResultType> {
  if(this.isDisconnected()){
    // We trust the connectionStatus to avoid useless request if we are offline
    return new Promise((accept) => accept({
      status: "timeout",
      response: null,
    }));
  }

  config = this.createConfiguration(config);
  this.callMiddlewares(config);

  return this.fetch(config).then(function(result : ResultType) {
    return new Result(result);
  }.bind(this));
};

RequestSender.interpretResult = interpretResult;
RequestSender.prototype.ContentTypes = ContentTypes;

export default RequestSenderFactory;