// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */
import objectMap from 'evojs/lib/objectMap';
import isArray from 'evojs/lib/isArray';
import isFixedNumber from 'evojs/lib/isFixedNumber';
import encodeObjectToURI from 'evojs/lib/encodeObjectToURI';
import { ContentTypes } from "../RequestSender";

import type { RequestConfiguration, ResponseType } from '../definitionTypes';
import type { FetchType } from '../RequestSender';
import { URITransformer, FormDataTransformer } from '../transformers';

const DEFAULT_TIMEOUT : number = 5000;

/**
 * Return a timeout if any, or by default return 5000
 * @param timeout
 * @returns {number}
 */
function getTimeoutDelay(timeout : any) : number {
  return isFixedNumber(timeout) ? timeout : DEFAULT_TIMEOUT;
}

/**
 * Map the headers into a map of headers with scalar values
 * @param headers
 */
function mapHeaders(headers : Object) : Object {
  return objectMap(headers, (value) => {
    if(isArray(value) && value.length === 1){
      return value[0];
    }

    return value;
  })
}

const TimeoutException = {
  name: "TimeoutException",
};

/**
 * make sure the promise rejects once the delay has expired
 * @param ms
 * @param promise
 * @returns {Promise}
 */
export const withTimeout = (ms : number, promise : Promise<*>) : Promise<*> => {
  return new Promise((resolve, reject) => {
    setTimeout(() => {
      reject(TimeoutException);
    }, ms);

    promise.then(resolve, reject);
  });
};

function urlAlreadyHasParameters(url){
  return url.indexOf("?") > -1;
}

function transformURLWithQuery(url, parameters){
  return urlAlreadyHasParameters(url) ? url + "&" + parameters : url + "?" + parameters;
}

/**
 *
 * @param response
 * @returns {*|Object}
 */
function getHeaders(response : ResponseType) {
  if(response.headers) {
    // Depending on the fetch implementation, headers are located in .map in natural fetch, _headers in node-fetch
    return response.headers.map ? mapHeaders(response.headers.map) : mapHeaders(response.headers._headers);
  }

  return {};
}

/**
 * Creates an adapter around the Fetch API
 * @param fetchFn
 */
const createAdapter = (fetchFn : Function) => {
  return (request : RequestConfiguration) => {
    const config : {
      method: string,
      headers: Object,
      body?: string,
    } = {
      method: request.method,
      headers: {},
    };

    if(request.headers){
      config.headers = request.headers;
    }

    // We override the Content-Type anyway
    if(request.body && typeof request.body === "object"){
      if(request.contentType === ContentTypes.formData){
        config.body = FormDataTransformer(request.body);
        config.headers["Content-Type"] = "multipart/form-data";
      } else if(request.contentType === ContentTypes.urlEncoded) {
        config.body = URITransformer(request.body);
        config.headers["Content-Type"] = "application/x-www-form-urlencoded";
      } else {
        config.body = JSON.stringify(request.body);
        config.headers["Content-Type"] = "application/json";
      }
    }

    // GET parameters must be passed in the URL
    let url = request.url;
    if(request.query){
      url = transformURLWithQuery(request.url, encodeObjectToURI(request.query));
    }

    return withTimeout(getTimeoutDelay(request.timeout), fetchFn(url, config))
      .then(response => {
        return response.json()
          .then(data => {
            return {
              status: "done",
              response: {
                statusCode: response.status,
                headers: getHeaders(response),
                data,
              }
            }
          })
          .catch((e) => {
            if(e.name === "SyntaxError"){
              return {
                status: "done",
                response: {
                  statusCode: response.status,
                  headers: getHeaders(response),
                  data: null,
                }
              }
            }
          })
      })
      .catch(error => {
        // See withTimeout
        if(error === TimeoutException){
          return {
            status: "timeout",
            response: null,
          }
        }

        return {
          status: "network-error",
          response: null,
        }
      });
  }
};

export default createAdapter;